﻿using System;
using System.Data;
using System.Web;
using System.Collections;
using System.Web.Services;
using System.Web.Services.Protocols;
using System.ComponentModel;
using System.Data.SqlClient;
using Microsoft.SqlServer.Types;
using System.Collections.Generic;
using System.Xml.Serialization;
using System.Text;
using System.Web.Script.Services;

namespace SpatialDemo
{
    // Don't forget to add these to the web.config at the end of system.web element
    /*
      <webServices>
        <protocols>
          <add name="HttpSoap"/>
          <add name="HttpPost"/>
          <add name="HttpGet"/>
        </protocols>
      </webServices>
    */


    [WebService(Namespace = "http://tempuri.org/")]
    [WebServiceBinding(ConformsTo = WsiProfiles.BasicProfile1_1)]
    [ToolboxItem(false)]
    // To allow this Web Service to be called from script, using ASP.NET AJAX, uncomment the following line. 
    [System.Web.Script.Services.ScriptService]
    public class Service1 : System.Web.Services.WebService
    {

        public static string connStr = @"Data Source=.;Initial Catalog=census;User ID=spatial_user;Password=StrongPW1;Connection Timeout=180;";
        public static readonly string[] spatialtables = new string[] { "zipcodes", "census", "congress" };

        [WebMethod]
        [ScriptMethod(UseHttpGet = true)]
        public string GeocodeAddress(string theaddress)
        {
            using (SqlConnection conn = new SqlConnection(connStr))
            using (SqlCommand cmd = new SqlCommand("select dbo.Geocode(theaddress)", conn))
            {
                string result = (string)cmd.ExecuteScalar();
                return result;
            }
        }


        //Id of the field that is: zipcode or census tract or congressional district #
        //s: source table 0 = zipcodes, 1 = census, 2 = congressional districts
        //i: use index 0=false, 1=true
        //t: include all regions that touch
        [WebMethod]
        [ScriptMethod(UseHttpGet=true)]
        public ShapeCollection GetShapeById(string id, int s, int t, int i)
        {
            using (SqlConnection sqlConn = new SqlConnection(connStr))
            {
                sqlConn.Open();
                using (SqlCommand cmd = new SqlCommand())
                {
                    cmd.Connection = sqlConn;
                    if (t == 1)
                        cmd.CommandText = String.Format("select convert(nvarchar(64), b.id), b.shape.STAsText() from {0} a, {0} b where a.id = @id and a.shape.STIntersects(b.shape)=1", spatialtables[s]);
                    else
                        cmd.CommandText = String.Format("SELECT convert(nvarchar(64), id), shape.STAsText() FROM {0} WHERE id = @id", spatialtables[s]);
                    cmd.Parameters.Add("@id", System.Data.SqlDbType.NVarChar, 5).Value = id;
                    SqlDataReader geomRdr = cmd.ExecuteReader();
                    SqlGeometry geom;
                    Shape shapePoints;
                    string name;
                    ShapeCollection shapeCollection = new ShapeCollection();
                    while (geomRdr.Read())
                    {
                        name = geomRdr.GetString(0);
                        geom = SqlGeometry.Parse(geomRdr.GetSqlString(1));
                        //Translate SqlGeometry into a Shape to add it to the ShapeCollection
                        for (int n = 0; n < geom.STNumGeometries(); n++)
                        {
                            SqlGeometry geomN = geom.STGeometryN(n + 1);
                            int numPoints = geomN.STNumPoints().Value;
                            shapePoints = new Shape(name, numPoints, 2);
                            for (int p = 0; p < numPoints; p++)
                            {
                                shapePoints.points[p] = new Point(geomN.STPointN(p + 1).STX.Value, geomN.STPointN(p + 1).STY.Value);
                            }
                            shapeCollection.shapeList.Add(shapePoints);
                        }
                    }
                    geomRdr.Close();
                    return shapeCollection;
                }
            }
        }

        // This method is used to get a shape from a table when the map is clicked
        // The parameter "s" is the table from which to fetch the data
        // The parameter "i" indicates whether or not the spatial index is used
        //                   for illustration purposes

        [WebMethod]
        [ScriptMethod(UseHttpGet = true)]
        public ShapeCollection GetShapeByLatLong(double lat, double lon, int s, int i)
        {
            SqlGeometry latlonPoint = SqlGeometry.Point(lat, lon, 0);
            using (SqlConnection sqlConn = new SqlConnection(connStr))
            {
                sqlConn.Open();
                using (SqlCommand cmd = new SqlCommand())
                {
                    cmd.CommandTimeout = 90;
                    cmd.Connection = sqlConn;

                    //Need to instantiate the object outside the STIntersects method
                    string indexHint;
                    if (i == 1)
                        indexHint = "idx_geom_" + spatialtables[s];
                    else
                        indexHint = "1";

                    cmd.CommandText = String.Format("select convert(nvarchar(64),a.id), a.shape.STAsText() from {0} a with (index({1}))where a.shape.STIntersects(@geom)=1 option(loop join)", spatialtables[s], indexHint);
                    cmd.Parameters.AddWithValue("@geom", latlonPoint);
                    cmd.Parameters[0].UdtTypeName = "geometry";
                    SqlDataReader geomRdr = cmd.ExecuteReader();
                    ShapeCollection shapeCollection = new ShapeCollection();
                    shapeCollection.shapeList = GeometryReaderToShape(geomRdr);
                    geomRdr.Close();
                    return shapeCollection;
                }
            }
        }

        /*
         * Subroutines to convert a SqlDataReader containing a spatial class
         * into a collection of shapes in XML format.
         * These shapes can be used as input to Virtual Earth layers
         * 
         */

        private List<Shape> GeometryReaderToShape(SqlDataReader geomRdr)
        {
            const double r = 3959; //Radius of the earth in miles
            int numFields = geomRdr.VisibleFieldCount;
            SqlGeometry pushpinLoc;
            StringBuilder pushpinData = new StringBuilder();
            SqlGeometry geom;
            Shape shapePoints;
            string name;
            string columnName;
            string columnValue;
            int color;
            double area;

            List<Shape> shapeList = new List<Shape>();

            while (geomRdr.Read())
            {
                name = geomRdr.GetValue(0).ToString().Replace("'", "");
                geom = SqlGeometry.Parse(geomRdr.GetSqlString(1));
                pushpinData.Length = 0;
                color = 1;
                for (int f = 2; f < numFields; f++)
                {
                    columnName = geomRdr.GetName(f);
                    columnValue = geomRdr.GetValue(f).ToString();
                    if (String.IsNullOrEmpty(columnValue))
                        columnValue = "0";
                    if (columnName == "area") //area column
                    {
                        area = double.Parse(columnValue);
                        area = Math.Round(r * Math.Cos(geom.STStartPoint().STX.Value * Math.PI / 180) * area, 2);
                        pushpinData.Append("Area: " + area.ToString() + " sq miles<br>");
                    }
                    //color column
                    else if (columnName == "color")
                        color = int.Parse(columnValue) - 1;
                    else
                        pushpinData.Append(columnName + ": " + columnValue.Replace("'", "") + "<br>");

                }
                int gtype = 0;
                for (int n = 0; n < geom.STNumGeometries(); n++)
                {
                    SqlGeometry geomN = geom.STGeometryN(n + 1);
                    int numPoints = geomN.STNumPoints().Value;
                    switch (geomN.STGeometryType().Value)
                    {
                        case "Point":
                            gtype = 0;
                            break;
                        case "LineString":
                            gtype = 1;
                            break;
                        case "Polygon":
                            gtype = 2;
                            break;
                        default:
                            throw new InvalidOperationException();
                    }
                    shapePoints = new Shape(name, numPoints, gtype);
                    if (geomN.STGeometryType() == "Polygon")
                        pushpinLoc = geomN.STCentroid();
                    else if (geomN.STGeometryType() == "Point")
                        pushpinLoc = geomN;
                    else //Linestring use first point for now
                        pushpinLoc = geomN.STPointN(1);
                    shapePoints.pushpin = new Point(pushpinLoc.STX.Value, pushpinLoc.STY.Value); ;
                    shapePoints.pushpinData = pushpinData.ToString();
                    shapePoints.color = color;
                    for (int p = 0; p < numPoints; p++)
                    {
                        shapePoints.points[p] = new Point(geomN.STPointN(p + 1).STX.Value, geomN.STPointN(p + 1).STY.Value);
                    }
                    shapeList.Add(shapePoints);
                }
            }
            return shapeList;

        }

        /*
         * Subroutines to convert a SqlDataReader containing a spatial class
         * into a collection of shapes in XML format.
         * These shapes can be used as input to Virtual Earth layers.
         * 
         * Note: the Geography class may not have all of the same methods as Geometry class
         * 
         */

        private List<Shape> GeographyReaderToShape(SqlDataReader geomRdr)
        {
            int numFields = geomRdr.VisibleFieldCount;
            SqlGeometry pushpinLoc;
            StringBuilder pushpinData = new StringBuilder();
            SqlGeometry geom;
            Shape shapePoints;
            string name;
            string columnName;
            string columnValue;
            int color;
            double area;
            int gtype;

            List<Shape> shapeList = new List<Shape>();

            while (geomRdr.Read())
            {
                name = geomRdr.GetValue(0).ToString().Replace("'", "");
                geom = SqlGeometry.Parse(geomRdr.GetSqlString(1));
                pushpinData.Length = 0;
                color = 0;
                for (int f = 2; f < numFields; f++)
                {
                    columnName = geomRdr.GetName(f);
                    columnValue = geomRdr.GetValue(f).ToString();
                    if (String.IsNullOrEmpty(columnValue))
                        columnValue = "0";
                    if (columnName == "area") //area column
                    {
                        area = double.Parse(columnValue);
                        pushpinData.Append("Area: " + area.ToString() + " sq meters<br>");
                    }
                    //color column
                    else if (columnName == "color")
                        color = int.Parse(columnValue) - 1;
                    else
                        pushpinData.Append(columnName + ": " + columnValue.Replace("'", "") + "<br>");

                }

                for (int n = 0; n < geom.STNumGeometries(); n++)
                {
                    SqlGeometry geomN = geom.STGeometryN(n + 1);
                    int numPoints = geomN.STNumPoints().Value;
                    switch (geomN.STGeometryType().Value)
                    {
                        case "Point":
                            gtype = 0;
                            break;
                        case "LineString":
                            gtype = 1;
                            break;
                        case "Polygon":
                            gtype = 2;
                            break;
                        default:
                            throw new InvalidOperationException();
                    }
                    shapePoints = new Shape(name, numPoints, gtype);
                    if (geomN.STGeometryType() == "Polygon")
                        pushpinLoc = geomN.STCentroid();
                    else if (geomN.STGeometryType() == "Point")
                        pushpinLoc = geomN;
                    else //Linestring use first point for now
                        pushpinLoc = geomN.STPointN(1);
                    shapePoints.pushpin = new Point(pushpinLoc.STX.Value, pushpinLoc.STY.Value); ;
                    shapePoints.pushpinData = pushpinData.ToString();
                    shapePoints.color = color;
                    for (int p = 0; p < numPoints; p++)
                    {
                        shapePoints.points[p] = new Point(geomN.STPointN(p + 1).STX.Value, geomN.STPointN(p + 1).STY.Value);
                    }
                    shapeList.Add(shapePoints);
                }
            }
            return shapeList;
        }

    }

}

public class ShapeCollection
{
    [XmlArrayAttribute(ElementName = "shapes")]
    [XmlArrayItem(ElementName = "shape", Type = typeof(Shape))]
    public List<Shape> shapeList = new List<Shape>();
}


public class Shape
{
    public Shape()
    {

    }

    public Shape(string name, int numPoints, int gtype)
    {
        this.name = name;
        this.points = new Point[numPoints];
        this.gtype = gtype;
        this.pushpin = null;
        this.pushpinData = null;
    }
    [XmlElement(ElementName = "name")]
    public string name;

    [XmlElement(ElementName = "gtype")]
    public int gtype; //0 = Point, 1=LineString, 2=Polygon

    [XmlElement(ElementName = "pushpin", Type = typeof(Point), IsNullable = true)]
    public Point pushpin;
    [XmlElement(ElementName = "pushpinData", IsNullable = true)]
    public string pushpinData;

    [XmlElement(ElementName = "color", IsNullable = false)]
    public int color;

    [XmlArrayAttribute(ElementName = "points")]
    [XmlArrayItem(ElementName = "point", Type = typeof(Point))]
    public Point[] points;
}

public class Point
{
    public Point()
    {

    }

    public Point(double x, double y)
    {
        this.x = x;
        this.y = y;
    }

    [XmlElement(ElementName = "x")]
    public double x;
    [XmlElement(ElementName = "y")]
    public double y;
}